#ifndef _PG_OSTREAM_HPP_
#define _PG_OSTREAM_HPP_

#include "boost/tuple/tuple.hpp"
#include <ostream>
#include <iomanip>

namespace ParGraph
{

template< class P >
std::ostream& operator<< ( std::ostream &os, const OwnerComputesGraph<P> &g )
{
  for( proc_t p = 0 ; p < size() ; p++ )
  {
    MPI_Barrier( _libbase.comm() );
    if( p == rank() )
    {
      typename OwnerComputesGraph<P>::const_VertexIterator vit, vend;
      boost::tie( vit, vend ) = g.get_local_vertices();
      for(; vit != vend ; vit++ )
      {
        os << "P" << _libbase.rank() << ":  ID " << std::setw(2) << std::right << get_id(*vit) << ": ";
	os << g.get_property_reference( *vit ) << std::endl;
      }
    }
  }
  MPI_Barrier( _libbase.comm() );
  return os;
}

template< class TAG, class TYPE, class NEXT >
std::ostream& operator<< ( std::ostream &os, const property<TAG,TYPE,NEXT> &prop )
{
  return os << std::boolalpha << get_property_value( prop, TAG() ) << ", " << (NEXT) prop;
}

template< class TAG, class TYPE >
std::ostream& operator<< ( std::ostream &os, const property<TAG,TYPE,no_property> &prop )
{
  return os << std::boolalpha << get_property_value( prop, TAG() );
}

std::ostream& operator<< ( std::ostream &os, default_color_type c )
{
  switch( c )
  {
    case black_color: return os << "black";
    case gray_color:  return os << "gray";
    case white_color: return os << "white";
  }
  return os;
}

template< class P >
std::ostream& operator<< ( std::ostream &os, const EdgeList<P,directed_out_edges_type> &edges )
{
  typename EdgeList<P,directed_out_edges_type>::const_mapiterator ebegin, eit, eend;
  boost::tie( ebegin, eend ) = out_edges( edges );
  std::cout << "-> { ";
  for( eit = ebegin ; eit != eend ; eit++ )
  {
    if( eit != ebegin ) std::cout << ", ";
    std::cout << eit->first << ";" << get_id(eit->second->get_v2()) << "( " << eit->second.get_property_reference() << ")";
  }
  return os << " }";
}

template< class P >
std::ostream& operator<< ( std::ostream &os, const EdgeList<P,directed_in_out_edges_type> &edges )
{
  typename EdgeList<P,directed_in_out_edges_type>::const_mapiterator ebegin, eit, eend;
  boost::tie( ebegin, eend ) = out_edges( edges );
  std::cout << "-> { ";
  for( eit = ebegin ; eit != eend ; eit++ )
  {
    if( eit != ebegin ) std::cout << ", ";
    std::cout << eit->first << ";" << get_id(eit->second.get_v2()) << "( " << eit->second.get_property_reference() << ")";
  }
  typename EdgeList<P,directed_in_out_edges_type>::const_mapiterator ebegin2, eit2, eend2;
  boost::tie( ebegin2, eend2 ) = in_edges( edges );
  std::cout << " } <- { ";
  for( eit2 = ebegin2 ; eit2 != eend2 ; eit2++ )
  {
    if( eit2 != ebegin2 ) std::cout << ", ";
    std::cout << eit2->first << ";" << get_id(eit2->second.get_v2()) << "( " << eit2->second.get_property_reference() << ")";
  }
  return os << " }";
}

std::ostream& operator<< ( std::ostream &os, const no_property& )
{
  return os;
}

}
#endif // _PG_OSTREAM_HPP_
