#ifndef _PG_PROPERTY_HPP_
#define _PG_PROPERTY_HPP_

#include <boost/pending/property.hpp>
#include <string>

namespace ParGraph
{
using boost::no_property;
using boost::property;

struct simple_distribution
{
  template< class Property, class Buffer >
  inline void put( const Property &prop, Buffer & buf ) const
  {
    buf.put( &prop, 1 );
  }

  template< class Property, class Buffer >
  inline void get( Property &prop, Buffer & buf ) const
  {
    buf.get( &prop, 1 );
  }
  template< class Tag, class Property, class Graph, class DistHelper, class Buffer >
  inline void put( Tag, const Property &prop, Graph &, DistHelper &, Buffer & buf ) const
  {
    buf.put( &prop, 1 );
  }

  template< class Tag, class Property, class Graph, class DistHelper, class Buffer >
  inline void get( Tag, Property &prop, Graph &, DistHelper &, Buffer & buf ) const
  {
    buf.get( &prop, 1 );
  }
};

template<class A>
struct property_distribution_traits
{
  typedef simple_distribution distribution_type;
};

struct string_distribution
{
  template< class chartype, class Buffer >
  inline void put( const std::basic_string<chartype> &prop, Buffer & buf ) const
  {
    size_t size = prop.size();
    buf.put( size );
    buf.put( prop.data(), size );
  }

  template< class chartype, class Buffer >
  inline void get( std::basic_string<chartype> &prop, Buffer & buf ) const
  {
    size_t size;
    buf.get( &size, 1 );
    if( size != 0 )
    {
      tmp = new chartype[ size + 1 ];
      buf.get( tmp, size );
      tmp[size] = 0;
      prop = tmp;
      delete[] tmp;
    }
  }

  template< class Tag, class chartype, class Graph, class DistHelper, class Buffer >
  inline void put( Tag, std::basic_string<chartype> &prop, Graph &, DistHelper &, Buffer & buf ) const
  {
    put( prop, buf );
  }

  template< class Tag, class chartype, class Graph, class DistHelper, class Buffer >
  inline void get( Tag, std::basic_string<chartype> &prop, Graph &, DistHelper &, Buffer & buf ) const
  {
    get( prop, buf );
  }
};

template<class T>
struct property_distribution_traits<std::basic_string<T> >
{
  typedef string_distribution distribution_type;
};

struct complex_distribution
{
  template< class Property, class Buffer >
  inline void put( const Property &prop, Buffer & buf )
  {
    prop._pgdist_fill_buf( buf );
  }

  template< class Property, class Buffer >
  inline void get( Property &prop, Buffer & buf )
  {
    prop._pgdist_read_buf( buf );
  }
  template< class Tag, class Property, class Graph, class DistHelper, class Buffer >
  inline void put( Tag t, Property &prop, Graph & g, DistHelper & dh, Buffer & buf )
  {
    prop._pgdist_fill_buf( t, g, dh, buf );
  }

  template< class Tag, class Property, class Graph, class DistHelper, class Buffer >
  inline void get( Tag t, Property &prop, Graph & g, DistHelper & dh, Buffer & buf )
  {
    prop._pgdist_read_buf( t, g, dh, buf );
  }
};

template< class PropertyList, class Buffer >
inline void
put_property_into_buffer( PropertyList& ref, Buffer & buf )
{
  typedef typename PropertyList::value_type value_type;
  typedef typename property_distribution_traits<value_type>::distribution_type distribution_type;

  distribution_type helper;

  helper.put( ref.m_value, buf );
  put_property_into_buffer( (typename PropertyList::next_type &) ref, buf );
}

template< class Buffer >
inline void
put_property_into_buffer( no_property&, Buffer & )
{
}

template< class Graph, class DistHelper, class PropertyList, class Buffer >
inline void
put_property_into_buffer( Graph & g, DistHelper & dh, PropertyList& ref, Buffer & buf )
{
  typedef typename PropertyList::value_type value_type;
  typedef typename property_distribution_traits<value_type>::distribution_type distribution_type;

  distribution_type helper;

  helper.put( typename PropertyList::tag_type(), ref.m_value, g, dh, buf );
  put_property_into_buffer( g, dh, (typename PropertyList::next_type &) ref, buf );
}

template< class Graph, class DistHelper, class Buffer >
inline void
put_property_into_buffer( Graph &, DistHelper &, no_property&, Buffer & )
{
}

template< class PropertyList, class Buffer >
inline void
get_property_from_buffer( PropertyList& ref, Buffer & buf )
{
  typedef typename PropertyList::value_type value_type;
  typedef typename property_distribution_traits<value_type>::distribution_type distribution_type;

  distribution_type helper;

  helper.get( ref.m_value, buf );
  get_property_from_buffer( (typename PropertyList::next_type &) ref, buf );
}

template< class Buffer >
inline void
get_property_from_buffer( no_property&, Buffer & )
{
}

template< class Graph, class DistHelper, class PropertyList, class Buffer >
inline void
get_property_from_buffer( Graph & g, DistHelper & dh, PropertyList& ref, Buffer & buf )
{
  typedef typename PropertyList::value_type value_type;
  typedef typename property_distribution_traits<value_type>::distribution_type distribution_type;

  distribution_type helper;

  helper.get( typename PropertyList::tag_type(), ref.m_value, g, dh, buf );
  get_property_from_buffer( g, dh, (typename PropertyList::next_type &) ref, buf );
}

template< class Graph, class DistHelper,  class Buffer >
inline void
get_property_from_buffer( Graph &, DistHelper &, no_property&, Buffer & )
{
}

typedef enum { white_color, gray_color, black_color } default_color_type;

template< class CT >
struct color_traits
{
  static default_color_type white() { return white_color; }
  static default_color_type gray() { return gray_color; }
  static default_color_type black() { return black_color; }
};

template<>
struct mpi_traits<default_color_type>
{
  const MPI_Datatype mpi_type;
  mpi_traits() : mpi_type(MPI_INT) {}
  // static const MPI_Datatype mpi_type = MPI_INT;
};

struct vertex_name_tag {};
struct vertex_color_tag {};
struct vertex_level_tag {};
struct vertex_proc_tag {};
struct vertex_potential_tag {};

struct edgelist1_tag {};
struct edgelist_residual_tag {};

struct edge_flow_tag {};
struct edge_capacity_tag {};
struct edge_reference_tag {};
struct edge_flag1_tag {};

}

#endif // _PG_PROPERTY_HPP_
